--[[
	Description: Announce various crowd-control effects on you.
	Author: Eliote
	This addon started when Canettieri and I couldn't find any like it that worked in Portuguese.
--]]

--local ADDON_NAME = "StunAlert"
--local VERSION = GetAddOnMetadata(ADDON_NAME, "Version")

local ADDON_NAME, L = ...
local format = string.format

local debugChatFrame

local STUN_ALERT = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0", "AceConsole-3.0")

local PLAYER_NAME = UnitName("player")
local DB = {}

local SPELLS = {}

local REGEX
REGEX = {
	match = function(name, desc)
		if not name then return false end

		for i, v in ipairs(REGEX.names) do
			if string.match(name, v) then return true end
		end

		for i, v in ipairs(REGEX.descriptions) do
			if string.match(desc, v) then return true end
		end

		return false
	end,
	names = L.NAMES or {},
	descriptions = L.DESCRIPTIONS or {
		"[fF]ear",
		"[sS]cared",
		"flee in terror",
		"[cC]harmed",
		"[iI]ncapacitat",
		"[dD]isoriented",
		"[sS]apped",
		"unable to act",
		"[cC]hoke",
		"[cC]hoking",
		"[dD]eep slumber",
		"[sS]lumber",
		"[aA]sleep",
		"[sS]leeping",
		"[rR]oot",
		"[fF]rozen",
		"[iI]mmobiliz",
		"[sS]ilence",
		"[cC]annot cast spells",
		"[pP]acified",
		"[sS]tun"
	}
}

local print = function(...)
	local out = ""

	for i = 1, select("#", ...) do
		local s = tostring(select(i, ...))

		out = out .. s
	end

	print("|cFF29E0E7<StunAlert>|r", out)
end

local dprint = function(...)
	local out = ""

	for i = 1, select("#", ...) do
		local s = tostring(select(i, ...))

		out = out .. s
	end

	local chat = (STUN_ALERT.CreateChatFrame and STUN_ALERT.CreateChatFrame()) or DEFAULT_CHAT_FRAME

	chat:AddMessage("|cFF29E0E7<StunAlert>|r" .. out)
end

local function interp(str, tab)
	return (str:gsub('($%b{})', function(w) return tostring(tab[w:sub(3, -2)] or w) end))
end

local SendChatMessage = function(msg, chat)
	if chat == "PRINT" then
		print(msg)
	else
		SendChatMessage(msg, chat)
	end
end

local function BoolValue(bool)
	if bool then return true else return false end
end

function STUN_ALERT.PrintDebug(event, spellId, icon)
	if not spellId then return end

	local spellLink = GetSpellLink(spellId)

	local name, _, _, _, debuffType, duration, expirationTime, unitCaster = STUN_ALERT.GetAura(spellId)

	if icon then
		icon = "|T" .. icon .. ":0|t"
	else
		icon = ""
	end

	dprint(icon, " ", event, ":{id:", spellId, ", spell:", spellLink, "}")
end

function STUN_ALERT.Enable(bool)
	DB.enable = BoolValue(bool)

	if DB.enable then
		STUN_ALERT:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	else
		STUN_ALERT:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end

	print(L["Enable"] .. ":", DB.enable)
end

function STUN_ALERT.Debug(int)
	DB.debugMode = int

	print(L["Debug"] .. ":", DB.debugMode)
end

function STUN_ALERT.Chat(chat)
	DB.chat = chat

	print(L["Chat"] .. ":", DB.chat)
end

function STUN_ALERT.Alert(spellId, sourceName, duration)
	local spellLink = GetSpellLink(spellId)

	if not sourceName then sourceName = "?" end
	if not spellLink then spellLink = "?" end

	duration = format(L["~%.1fs"], duration)

	SendChatMessage(interp(DB.startText, { source = sourceName, skill = spellLink, player = PLAYER_NAME, remainingTime = duration }), DB.chat)
end

function STUN_ALERT.AlertEnd(spellId, sourceName)
	local spellLink = GetSpellLink(spellId)

	if not sourceName then sourceName = "?" end
	if not spellLink then spellLink = "?" end

	SendChatMessage(interp(DB.endText, { source = sourceName, skill = spellLink, player = PLAYER_NAME }), DB.chat)
end

function STUN_ALERT.GetAura(id)
	for i = 1, 40 do
		local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitAura("player", i, "HARMFUL")

		if name then
			if id == spellId then
				return name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate
			end
		else
			return nil
		end
	end
end

function STUN_ALERT.GetAuraDuration(expirationTime)
	local remain = expirationTime - GetTime()

	return math.floor(remain)
end

function STUN_ALERT:OnInitialize()
	local defaults = {
		profile = {
			enable = true,
			debugMode = 0,
			chat = "PARTY",
			startText = L["{rt7}SURPRISE!!! ${source} use ${skill} on ${player} (${remainingTime} remaining){rt7}"],
			endText = L["{rt6}${skill} ended!{rt6}"],
			minDuration = 1,
			printBroken = true,
			useDebugWindow = true,
			debugChatFrame = -1
		}
	}
	self.db = LibStub("AceDB-3.0"):New("StunAlertDB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "ReloadOptions")
	self.db.RegisterCallback(self, "OnProfileCopied", "ReloadOptions")
	self.db.RegisterCallback(self, "OnProfileReset", "ReloadOptions")

	STUN_ALERT:ReloadOptions()

	local options = {
		name = ADDON_NAME,
		type = "group",
		args = {
			enable = {
				type = "toggle",
				name = L["Enable"],
				desc = L["Enable/Disable the addon"],
				order = 1,
				width = "full",
				set = function(info, val) STUN_ALERT.Enable(val) end,
				get = function(info) return DB.enable end
			},
			printBroken = {
				type = "toggle",
				name = L["Warn end"],
				desc = L["Enable/Disable warning when CC ends"],
				order = 2,
				width = "full",
				set = function(info, val) DB.printBroken = val end,
				get = function(info) return DB.printBroken end
			},
			debug = {
				type = "select",
				name = L["Debug"],
				desc = L["Debug level"],
				order = 3,
				values = {
					[0] = "0",
					[1] = "1",
					[2] = "2",
					[3] = "3",
					[4] = "4",
					[5] = "5"
				},
				set = function(info, val) STUN_ALERT.Debug(val) end,
				get = function(info) return DB.debugMode end
			},
			spacer1 = {
				type = "description",
				name = "", --use an empty value for name
				order = 4,
				width = "full", --forces a new line after the items above it
			},
			useDebugWindow = {
				type = "toggle",
				name = L["Use a new chat for debug"],
				order = 5,
				width = "full",
				set = function(info, val) DB.useDebugWindow = val end,
				get = function(info) return DB.useDebugWindow end
			},
			debugChatFrame = {
				type = "range",
				name = L["Chat frame"],
				order = 5.1,
				min = 3,
				max = 10,
				step = 1,
				set = function(info, val) DB.debugChatFrame = val; if debugChatFrame then FCF_Close(debugChatFrame) end; debugChatFrame = nil  end,
				get = function(info) return DB.debugChatFrame end
			},
			spacer1 = {
				type = "description",
				name = "", --use an empty value for name
				order = 5.9,
				width = "full", --forces a new line after the items above it
			},
			chat = {
				type = "select",
				name = L["Chat"],
				desc = L["Chat to announce the CC"],
				order = 6,
				values = {
					["SAY"] = L["Say"],
					["PARTY"] = L["Party"],
					["GUILD"] = L["Guild"],
					["OFFICER"] = L["Officer"],
					["RAID"] = L["Raid"],
					["RAID_WARNING"] = L["Raid Warning"],
					["INSTANCE_CHAT"] = L["Instance"],
					["YELL"] = L["Yell"],
					["BATTLEGROUND"] = L["Battleground"],
					["PRINT"] = L["Print"]
				},
				set = function(info, val) STUN_ALERT.Chat(val) end,
				get = function(info) return DB.chat end
			},
			spacer2 = {
				type = "description",
				name = "", --use an empty value for name
				order = 7,
				width = "full", --forces a new line after the items above it
			},
			minDuration = {
				type = "range",
				name = L["Minimal duration to alert"],
				order = 8,
				min = 0,
				max = 10,
				step = 1,
				set = function(info, val) DB.minDuration = val end,
				get = function(info) return DB.minDuration end
			},
			textStart = {
				type = "input",
				name = L["CC start text"],
				order = 9,
				width = "full",
				set = function(info, val) DB.startText = val end,
				get = function(info) return DB.startText end
			},
			textEnd = {
				type = "input",
				name = L["CC end text"],
				order = 10,
				width = "full",
				set = function(info, val) DB.endText = val end,
				get = function(info) return DB.endText end
			}
		},
	}

	local AceConfig = LibStub("AceConfig-3.0")
	AceConfig:RegisterOptionsTable(ADDON_NAME, options, { "stun" })
	AceConfig:RegisterOptionsTable("StunAlert-Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db))

	local AceDialog = LibStub("AceConfigDialog-3.0")
	AceDialog:AddToBlizOptions(ADDON_NAME, "StunAlert")
	AceDialog:AddToBlizOptions("StunAlert-Profiles", "Profiles", ADDON_NAME)
end

function STUN_ALERT:ReloadOptions()
	DB = self.db.profile

	STUN_ALERT.Enable(DB.enable)
end

function STUN_ALERT.CreateChatFrame()
	if not DB.useDebugWindow then return end
	if debugChatFrame then return debugChatFrame end

	local id = DB.debugChatFrame

	local chatFrame = _G["ChatFrame" .. id];
	local chatTab = _G["ChatFrame" .. id .. "Tab"];

	-- initialize the frame
	FCF_SetWindowName(chatFrame, "Debug");
	FCF_SetWindowColor(chatFrame, DEFAULT_CHATFRAME_COLOR.r, DEFAULT_CHATFRAME_COLOR.g, DEFAULT_CHATFRAME_COLOR.b);
	FCF_SetWindowAlpha(chatFrame, DEFAULT_CHATFRAME_ALPHA);
	SetChatWindowLocked(id, nil);

	-- clear stale messages
	chatFrame:Clear();

	-- Listen to the standard messages
	ChatFrame_RemoveAllMessageGroups(chatFrame);
	ChatFrame_RemoveAllChannels(chatFrame);

	--Clear the edit box history.
	chatFrame.editBox:ClearHistory();

	-- Show the frame and tab
	chatFrame:Show();
	chatTab:Show();
	--SetChatWindowShown(id, 1);

	FCF_DockFrame(chatFrame, (#FCFDock_GetChatFrames(GENERAL_CHAT_DOCK)+1), true);
	FCF_FadeInChatFrame(FCFDock_GetSelectedWindow(GENERAL_CHAT_DOCK));

	debugChatFrame = chatFrame
	return chatFrame
end

function STUN_ALERT.Debugger(mode, eventType, spellId)
	local self = {
		debug = mode > 0
	}

	function self.printIf(n, ico)
		if self.debug and n <= mode then
			if not ico then ico = select(3, GetSpellInfo(spellId)) end

			STUN_ALERT.PrintDebug(eventType, spellId, ico)
			self.debug = false
		end
	end

	return self
end

function STUN_ALERT.COMBAT_LOG_EVENT_UNFILTERED(self, event, timestamp, eventType, hideCaster,
												sourceGUID, sourceName, sourceFlags, sourceRaidFlags,
												destGUID, destName, destFlags, destRaidFlags, spellId,
												spellName, spellSchool)
	if not spellId then return end

	local debugger = STUN_ALERT.Debugger(DB.debugMode, eventType, spellId)
	debugger.printIf(5)

	if destName ~= PLAYER_NAME then return end

	if eventType == "SPELL_AURA_APPLIED" or eventType == "SPELL_AURA_REFRESH" then
		debugger.printIf(3)

		local name, _, icon, count, debuffType, duration, expirationTime = STUN_ALERT.GetAura(spellId)
		if not name then return end

		debugger.printIf(1, icon)

		if not SPELLS[spellId] then
			local description = GetSpellDescription(spellId) or ''

			if not REGEX.match(name, description) then return end
		end

		SPELLS[spellId] = duration

		local duration = STUN_ALERT.GetAuraDuration(expirationTime)
		if duration <= DB.minDuration then return end

		STUN_ALERT.Alert(spellId, sourceName, duration)
	elseif eventType == "SPELL_AURA_BROKEN" or eventType == "SPELL_AURA_BROKEN_SPELL" or eventType == "SPELL_AURA_REMOVED" then
		debugger.printIf(4)

		if not DB.printBroken then return end
		if not SPELLS[spellId] then return end
		if SPELLS[spellId] <= DB.minDuration then return end

		debugger.printIf(2)

		STUN_ALERT.AlertEnd(spellId, sourceName)
	end
end


